import os
from random import shuffle
from typing import Optional
import numpy as np
from bigfish import stack, detection
from skimage.morphology import extrema, opening
from scipy import ndimage
from scipy.optimize import linear_sum_assignment
import matplotlib as mpl

from cnn_framework.utils.display_tools import display_progress

from ..constants.tracking import CYTOKINESIS_DURATION
from ..utils.mid_body_track import MidBodyTrack
from ..utils.image_tools import smart_cropping
from ..utils.mid_body_spot import MidBodySpot
from ..utils.mitosis_track import MitosisTrack
from ..utils.trackmate_track import TrackMateTrack
from ..utils.tools import plot_detection


class MidBodyDetectionFactory:
    """
    Class to perform mid-body detection, tracking and filtering.

    Args:
        weight_mklp_intensity_factor (float): Weight of intensity in spot dist calculation
            (cf TrackMate).
        weight_sir_intensity_factor (float): Weight of sir intensity in spot distance calculation.
        mid_body_linking_max_distance (int): Maximum distance between two mid-bodies to link them.

        h_maxima_threshold (float): Threshold for h_maxima detection (default).

        sigma (float): Sigma for bigfish detection (unused).
        threshold (float): Threshold for bigfish detection (unused).

        cytokinesis_duration (int): Number of frames to look for mid-body in between cells.
        minimum_mid_body_track_length (int): Minimum spots in mid-body track to consider it.
    """

    def __init__(
        self,
        weight_mklp_intensity_factor=10.0,
        weight_sir_intensity_factor=3.33,
        mid_body_linking_max_distance=100,
        h_maxima_threshold=5.0,
        sigma=2.0,
        threshold=1.0,
        cytokinesis_duration=CYTOKINESIS_DURATION,
        minimum_mid_body_track_length=10,
    ) -> None:
        self.weight_mklp_intensity_factor = weight_mklp_intensity_factor
        self.weight_sir_intensity_factor = weight_sir_intensity_factor
        self.mid_body_linking_max_distance = mid_body_linking_max_distance
        self.h_maxima_threshold = h_maxima_threshold
        self.sigma = sigma
        self.threshold = threshold
        self.cytokinesis_duration = cytokinesis_duration
        self.minimum_mid_body_track_length = minimum_mid_body_track_length

    def update_mid_body_spots(
        self,
        mitosis_track: MitosisTrack,
        mitosis_movie: np.array,
        mask_movie: np.array,
        tracks: list[TrackMateTrack],
    ) -> None:
        """
        Get spots of best mitosis track.

        Parameters
        ----------
        mitosis_movie: TYXC
        mask_movie: TYX

        """

        spots_candidates = self.detect_mid_body_spots(
            mitosis_movie, mask_movie=mask_movie
        )
        mid_body_tracks = self.generate_tracks_from_spots(spots_candidates)
        kept_track = self._select_best_track(
            mitosis_track, mid_body_tracks, tracks, mitosis_movie
        )

        if kept_track is None:
            return

        # Keep only spots of best mitosis track
        for rel_frame, spot in kept_track.spots.items():
            frame = rel_frame + mitosis_track.min_frame
            mitosis_track.mid_body_spots[frame] = spot

    def detect_mid_body_spots(
        self,
        mitosis_movie: np.array,
        mask_movie: Optional[np.array] = None,
        mid_body_channel=1,
        sir_channel=0,
        mode="h_maxima",
    ) -> dict[int, list[MidBodySpot]]:
        """
        Parameters
        ----------
        mitosis_movie: TYXC
        mask_movie: TYX

        Returns
        ----------
        spots_dictionary: dict[int, list[MidBodySpot]]
        """

        # Default mask is all ones
        if mask_movie is None:
            mask_movie = np.ones(mitosis_movie.shape[:-1])

        # Detect spots in each frame
        spots_dictionary = {}
        nb_frames = mitosis_movie.shape[0]
        for frame in range(nb_frames):
            display_progress(
                "Detect mid-body spots...",
                frame + 1,
                nb_frames,
                additional_message=f"Frame {frame + 1}/{nb_frames}",
            )

            mitosis_frame = mitosis_movie[frame, :, :, :].squeeze()  # YXC
            mask_frame = mask_movie[frame, :, :].squeeze()  # YX
            spots = self._spot_detection(
                mitosis_frame,
                mask_frame,
                mid_body_channel,
                sir_channel,
                mode=mode,
                frame=frame,
            )

            # Update dictionary
            spots_dictionary[frame] = spots

        return spots_dictionary

    def _spot_detection(
        self,
        image: np.array,
        mask: np.array,
        mid_body_channel: int,
        sir_channel: int,
        mode: str,
        frame=-1,
    ) -> list[MidBodySpot]:
        """
        Mode 'bigfish'
            threshold_1: sigma for log filter
            threshold_2: threshold for spots detection

        Mode 'h_maxima'
            threshold_1: threshold for h_maxima
            threshold_2: unused

        """

        image_sir = image[:, :, sir_channel]
        image_mklp = image[:, :, mid_body_channel]  #

        if mode == "bigfish":
            # Spots detection with bigfish functions
            filtered_image = stack.log_filter(image_mklp, sigma=self.sigma)
            # Filter out spots which are not maximal or outside convex hull
            spots_mask = (filtered_image > 0) * mask
            # If mask is empty, skip frame
            if np.sum(spots_mask) == 0:
                spots = np.array([], dtype=np.int64).reshape(
                    (0, filtered_image.ndim)
                )
            else:
                spots, _ = detection.spots_thresholding(
                    filtered_image,
                    spots_mask.astype(bool),
                    threshold=self.threshold,
                )

        elif mode == "h_maxima":
            # Perform opening followed by closing to remove small spots
            filtered_image = opening(image_mklp, footprint=np.ones((3, 3)))
            # Get local maxima using h_maxima
            local_maxima = extrema.h_maxima(
                filtered_image, self.h_maxima_threshold
            )
            # Label spot regions
            labeled_local_maxima, nb_labels = ndimage.label(
                local_maxima, structure=np.ones((3, 3))
            )
            # Remove inconsistent labels
            # Threshold is computed as 99th percentile of image
            filtering_threshold = np.quantile(image_mklp.flatten(), 0.99)
            for label in range(1, nb_labels + 1):
                # Labels intensity in original image has to be higher than threshold
                if (
                    image_mklp[np.where(labeled_local_maxima == label)].mean()
                    < filtering_threshold
                ):
                    labeled_local_maxima[labeled_local_maxima == label] = 0
            # Re-label accordingly
            labeled_local_maxima, nb_labels = ndimage.label(
                labeled_local_maxima > 0, structure=np.ones((3, 3))
            )
            # Get center of mass to locate spots
            spots = ndimage.center_of_mass(
                local_maxima, labeled_local_maxima, range(1, nb_labels + 1)
            )
            spots = np.asarray(spots, dtype=np.int64)

            # Here, do something to retrieve mid_body area and/or circularity...

            if len(spots) == 0:
                spots = np.array([], dtype=np.int64).reshape(
                    (0, filtered_image.ndim)
                )

        else:
            raise ValueError(f"Unknown mode: {mode}")

        # Convert spots to MidBodySpot objects (switch (y, x) to (x, y))
        mid_body_spots = [
            MidBodySpot(
                frame,
                x=position[1],
                y=position[0],
                intensity=self._get_average_intensity(position, image_mklp),
                sir_intensity=self._get_average_intensity(position, image_sir),
            )
            for position in spots
        ]

        return mid_body_spots

    @staticmethod
    def _get_average_intensity(
        position: tuple[int], image: np.array, margin=1
    ) -> int:
        """
        Parameters
        ----------
        position: (y, x)
        image: YX
        margin: int

        Returns
        ----------
        average_intensity: int
        """
        # Get associated crop
        crop = smart_cropping(
            image,
            margin,
            position[1],
            position[0],
            position[1] + 1,
            position[0] + 1,
        )

        # Return average intensity
        return int(np.mean(crop))

    def _update_spots_hereditary(
        self, spots1: list[MidBodySpot], spots2: list[MidBodySpot]
    ) -> None:
        """
        Link spots together using Hungarian algorithm.
        """
        # Ignore empty spots list
        if len(spots1) == 0 or len(spots2) == 0:
            return

        # Create cost matrix
        # https://imagej.net/plugins/trackmate/algorithms
        cost_matrix = np.zeros(
            (len(spots1) + len(spots2), len(spots1) + len(spots2))
        )
        max_cost = 0
        for i, spot1 in enumerate(spots1):
            for j, spot2 in enumerate(spots2):
                intensity_penalty = (
                    3
                    * self.weight_mklp_intensity_factor
                    * np.abs(spot1.intensity - spot2.intensity)
                    / (spot1.intensity + spot2.intensity)
                )
                sir_intensity_penalty = (
                    3
                    * self.weight_sir_intensity_factor
                    * np.abs(spot1.sir_intensity - spot2.sir_intensity)
                    / (spot1.sir_intensity + spot2.sir_intensity)
                )
                penalty = 1 + intensity_penalty + sir_intensity_penalty
                distance = spot1.distance_to(spot2)
                if distance > self.mid_body_linking_max_distance:
                    cost_matrix[i, j] = np.inf
                else:
                    # Compared to original TrackMate algorithm, remove square to penalize no attribution to the closest spot
                    cost_matrix[i, j] = (penalty * distance) ** 1
                    max_cost = max(max_cost, cost_matrix[i, j])

        min_cost = (
            0
            if np.max(cost_matrix) == 0
            else np.min(cost_matrix[np.nonzero(cost_matrix)])
        )

        cost_matrix[len(spots1) :, : len(spots2)] = (
            max_cost * 1.05
        )  # bottom left
        cost_matrix[: len(spots1), len(spots2) :] = (
            max_cost * 1.05
        )  # top right
        cost_matrix[len(spots1) :, len(spots2) :] = min_cost  # bottom right

        # Hungarian algorithm
        row_ind, col_ind = linear_sum_assignment(cost_matrix)

        # Update parent and child spots
        for i, j in zip(row_ind, col_ind):
            if i < len(spots1) and j < len(spots2):
                spots1[i].child_spot = spots2[j]
                spots2[j].parent_spot = spots1[i]

    def generate_tracks_from_spots(
        self, spots_candidates: dict[int, list[MidBodySpot]]
    ) -> list[MidBodyTrack]:
        """
        Use spots linked together to generate tracks.

        Parameters
        ----------
        spots_candidates : {frame: [MidBodySpot]}

        Returns
        ----------
        tracks: [MidBodyTrack]

        """

        # Update parent and child spots
        for frame in spots_candidates.keys():
            # Ignore if no spot detected in next frame
            if not frame + 1 in spots_candidates:
                continue
            self._update_spots_hereditary(
                spots_candidates[frame], spots_candidates[frame + 1]
            )

        tracks = []
        for spots in spots_candidates.values():
            for spot in spots:
                if spot.track_id is None:
                    # Create new track
                    track_id = len(tracks)
                    new_track = MidBodyTrack(track_id)
                    new_track.add_spot(spot)
                    tracks.append(new_track)

        return tracks

    def _select_best_track(
        self,
        mitosis_track: MitosisTrack,
        mid_body_tracks: list[MidBodyTrack],
        trackmate_tracks: list[TrackMateTrack],
        mitosis_movie: np.array,
        sir_channel=0,
    ) -> MidBodyTrack:
        """
        Select best track from mid-body tracks.

        Parameters
        ----------
        mitosis_movie: TYXC
        """
        (
            mother_track,
            daughter_tracks,
        ) = mitosis_track.get_mother_daughters_tracks(trackmate_tracks)
        # NB: only first daughter is considered
        daughter_track = daughter_tracks[0]

        expected_positions = {}
        for frame in range(
            daughter_track.start,
            daughter_track.start + self.cytokinesis_duration,
        ):
            # If one cell does not exist anymore, stop
            if (
                frame not in daughter_track.spots
                or frame not in mother_track.spots
            ):
                continue
            # Compute mid-body expected relative position at current frame
            closest_points = []
            min_distance = np.inf
            for mother_point in mother_track.spots[frame].spot_points:
                position_mother = [
                    int(mother_point[0]) - mitosis_track.position.min_x,
                    int(mother_point[1]) - mitosis_track.position.min_y,
                ]
                for daughter_point in daughter_track.spots[frame].spot_points:
                    position_daughter = [
                        int(daughter_point[0]) - mitosis_track.position.min_x,
                        int(daughter_point[1]) - mitosis_track.position.min_y,
                    ]
                    distance = np.linalg.norm(
                        [
                            a - b
                            for a, b in zip(position_mother, position_daughter)
                        ]
                    )
                    if distance < min_distance:
                        min_distance = distance
                        closest_points = [(position_mother, position_daughter)]
                    if distance == min_distance:
                        closest_points.append(
                            (position_mother, position_daughter)
                        )

            mid_body_position = np.mean(closest_points, axis=0)
            mid_body_position = np.mean(mid_body_position, axis=0)
            expected_positions[frame - mitosis_track.min_frame] = (
                mid_body_position
            )

        # Remove wrong tracks by keeping only tracks with at least minimum_track_length points
        mid_body_tracks = [
            track
            for track in mid_body_tracks
            if track.length > self.minimum_mid_body_track_length
        ]

        # Compute mean intensity on sir-tubulin channel for each track
        image_sir = mitosis_movie[..., sir_channel]  # TYX
        sir_intensity_track = [0 for _ in mid_body_tracks]
        for idx, track in enumerate(mid_body_tracks):
            abs_track_frames = [
                frame + mitosis_track.min_frame
                for frame in list(track.spots.keys())
            ]
            abs_min_frame = mitosis_track.key_events_frame[
                "cytokinesis"
            ]  # Cytokinesis start
            abs_max_frame = abs_min_frame + int(self.cytokinesis_duration / 2)
            if (
                abs_min_frame > abs_track_frames[-1]
                or abs_max_frame < abs_track_frames[0]
            ):
                sir_intensity_track[idx] = -np.inf
            frame_count = 0
            for frame in range(abs_min_frame, abs_max_frame + 1):
                if frame not in abs_track_frames:
                    continue
                frame_count += 1
                track_spot = track.spots[frame - mitosis_track.min_frame]
                sir_intensity_track[idx] += image_sir[
                    frame - mitosis_track.min_frame,
                    track_spot.y,
                    track_spot.x,
                ]

            if frame_count < (abs_max_frame - abs_min_frame + 1) / 2:
                sir_intensity_track[idx] = -np.inf
            else:
                sir_intensity_track[idx] /= frame_count

        # Get list of expected distances
        expected_distances = []
        for track in mid_body_tracks:
            val = track.get_expected_distance(
                expected_positions, self.mid_body_linking_max_distance
            )
            expected_distances.append(val)

        # Assert lists have same length for next function
        assert len(expected_distances) == len(mid_body_tracks)
        assert len(sir_intensity_track) == len(mid_body_tracks)

        # function to sort tracks by expected distance and intensity
        def func_sir_intensity(track):
            a = expected_distances[mid_body_tracks.index(track)]
            b = sir_intensity_track[mid_body_tracks.index(track)]
            return a - 0.5 * b

        # Remove tracks with infinite func value
        fun_values = []
        final_tracks = []
        for track in mid_body_tracks:
            fun_values.append(func_sir_intensity(track))
            if func_sir_intensity(track) != np.inf:
                final_tracks.append(track)

        # Sort tracks by func value
        sorted_tracks = sorted(final_tracks, key=func_sir_intensity)
        return sorted_tracks[0] if len(sorted_tracks) > 0 else None

    def save_mid_body_tracking(
        self,
        spots_candidates,
        mitosis_movie: np.ndarray,
        path_output: str,
        mid_body_channel=1,
    ):
        """
        Plot spots detection & tracking.
        """
        # Check if directory exists
        if not os.path.exists(path_output):
            os.makedirs(path_output)

        matplotlib_colors = [
            mpl.colormaps["hsv"](i)[:3] for i in np.linspace(0, 0.9, 100)
        ]
        shuffle(matplotlib_colors)

        # Detect spots in each frame
        nb_frames = mitosis_movie.shape[0]
        for frame in range(nb_frames):
            image = mitosis_movie[frame, :, :, mid_body_channel].squeeze()

            # Bigfish spots
            frame_spots = [
                [spot.y, spot.x] for spot in spots_candidates[frame]
            ]
            colors = [
                (
                    matplotlib_colors[spot.track_id % len(matplotlib_colors)]
                    if spot.track_id is not None
                    else (0, 0, 0)
                )
                for spot in spots_candidates[frame]
            ]

            plot_detection(
                image,
                frame_spots,
                color=colors,
                contrast=True,
                path_output=os.path.join(
                    path_output, f"spot_detection_{frame}.png"
                ),
                show=False,
                title=f"Python frame {frame} - Fiji frame {frame + 1}",
                fill=True,
            )



from __future__ import annotations

import numpy as np


class Spot:
    """
    Class used for both cell and mid-body spots.
    """

    def __init__(self, frame: int, x: int, y: int):
        self.frame = frame
        self.x = x
        self.y = y

    def get_position(self) -> np.ndarray:
        """
        Return position as numpy array
        """
        return np.array([self.x, self.y])

    def distance_to(self, other_spot: Spot):
        """
        Compute distance between two spots
        """
        return np.linalg.norm(
            np.array(self.get_position()) - np.array(other_spot.get_position())
        )



import numpy as np
import albumentations as A

from cnn_framework.utils.data_sets.abstract_data_set import AbstractDataSet
from cnn_framework.utils.data_sets.dataset_output import DatasetOutput
from cnn_framework.utils.tools import handle_image_type


class CnnDataSet(AbstractDataSet):
    """
    Custom class to avoid loading images from folder.
    """

    def __init__(self, data, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = data

    def set_transforms(self):
        height, width = self.params.input_dimensions.to_tuple(False)
        self.transforms = A.Compose(
            [
                A.Normalize(
                    self.mean_std["mean"],
                    std=self.mean_std["std"],
                    max_pixel_value=1,
                ),
                A.PadIfNeeded(
                    min_height=height,
                    min_width=width,
                    border_mode=0,
                    value=0,
                    p=1,
                ),
                A.CenterCrop(height=height, width=width, p=1),
            ]
        )

    def generate_images(self, filename):
        idx = int(filename.split(".")[0])
        # Get image and adapt it to torch
        nucleus_image = np.moveaxis(self.data[idx], 0, -1)  # YXC
        nucleus_image = handle_image_type(nucleus_image)  # to [0, 1]
        # Define any target value
        target_array = np.zeros(self.params.nb_classes)
        target_array[0] = 1
        # Construct output
        return DatasetOutput(
            input=nucleus_image,
            target_array=target_array,
        )



from .cell_spot import CellSpot


class TrackMateSpot(CellSpot):
    """
    Parse TrackMate spot from xml file.
    """

    def __init__(self, trackmate_spot, raw_video_shape: list[int]):
        x = float(trackmate_spot["@POSITION_X"])
        y = float(trackmate_spot["@POSITION_Y"])
        frame = int(trackmate_spot["@FRAME"])
        id_number = int(trackmate_spot["@ID"])

        # Get min and max positions
        raw_positions: str = trackmate_spot["#text"].split(" ")
        rel_positions_x = [
            float(raw_positions[i]) for i in range(0, len(raw_positions), 2)
        ]
        rel_positions_y = [
            float(raw_positions[i]) for i in range(1, len(raw_positions), 2)
        ]

        rel_min_x, rel_max_x = min(rel_positions_x), max(rel_positions_x)
        rel_min_y, rel_max_y = min(rel_positions_y), max(rel_positions_y)

        # Get spot points
        positions_x = [
            int(x + float(raw_positions[i]))
            for i in range(0, len(raw_positions), 2)
        ]
        positions_y = [
            int(y + float(raw_positions[i]))
            for i in range(1, len(raw_positions), 2)
        ]
        spot_points = [[x, y] for x, y in zip(positions_x, positions_y)]

        # Get min and max positions
        abs_min_x, abs_max_x = (
            int(x + rel_min_x),
            int(x + rel_max_x),
        )
        abs_min_y, abs_max_y = (
            int(y + rel_min_y),
            int(y + rel_max_y),
        )

        # Clip to video size
        abs_min_x, abs_max_x = (
            max(abs_min_x, 0),
            min(abs_max_x, raw_video_shape[2]),
        )
        abs_min_y, abs_max_y = (
            max(abs_min_y, 0),
            min(abs_max_y, raw_video_shape[1]),
        )

        super().__init__(
            frame,
            x,
            y,
            id_number,
            abs_min_x,
            abs_max_x,
            abs_min_y,
            abs_max_y,
            spot_points,
        )



import os
from cut_detector.data.tools import get_data_path
from cut_detector.utils.tools import upload_annotations


def test_annotations_upload(
    annotations_folder=get_data_path("annotations"),
    video_folder=get_data_path("videos"),
    mitoses_folder=get_data_path("mitoses"),
):
    # Initialize variables
    all_detected, all_not_detected = 0, 0

    video_paths = os.listdir(video_folder)
    for video_idx, video in enumerate(video_paths):
        video_progress = f"{video_idx + 1}/{len(video_paths)}"
        print(f"\nVideo {video_progress}...")
        # Perform mid-body detection evaluation
        detected, not_detected = upload_annotations(
            annotations_folder,
            os.path.join(video_folder, video),
            mitoses_folder,
            update_mitoses=False,
        )
        all_detected += detected
        all_not_detected += not_detected

    assert all_not_detected == 0



import os
import urllib.request

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


def get_data_path(data_name: str) -> None:
    """
    Returns absolute path to model file.
    """
    sub_folder_to_create = data_name
    if data_name == "mitoses":
        files = ["example_video_mitosis_0_4_to_0.bin"]
    elif data_name == "models":
        files = ["example_video_model.xml"]
    elif data_name == "tracks":
        sub_folder_to_create = f"{data_name}/example_video"
        files = [
            f"example_video/track_{track_id}.bin" for track_id in range(5)
        ]
    elif data_name == "results":
        files = []  # no files to download here
    elif data_name == "videos":
        files = ["example_video.tif"]
    elif data_name == "mitosis_movies":
        files = ["example_video_mitosis_0_0_to_4.tiff"]
    elif data_name == "mid_bodies":
        files = []  # no files to download here
    elif data_name == "segmentation_results":
        files = ["example_video.bin"]
    elif data_name == "spots":
        sub_folder_to_create = f"{data_name}/example_video"
        files = [f"example_video/spot_{spot_id}.bin" for spot_id in range(237)]
    elif data_name == "annotations":
        sub_folder_to_create = f"{data_name}/example_video"
        files = [
            "example_video/CellCounter_example_video_mitosis_0_0_to_4.xml"
        ]
    else:
        raise ValueError(f"Unknown data name: {data_name}")

    local_path = os.path.join(CURRENT_DIR, data_name)
    os.makedirs(os.path.join(CURRENT_DIR, sub_folder_to_create), exist_ok=True)

    for file in files:
        file_local_path = os.path.join(local_path, file)
        if not os.path.exists(file_local_path):
            print(f"Downloading data {data_name}...")
            urllib.request.urlretrieve(
                f"https://raw.githubusercontent.com//15bonte/cut-detector-models/main/data/{data_name}/{file}",
                file_local_path,
            )

    return local_path



import json
import os
import numpy as np
import xmltodict

from ..utils.cell_track import CellTrack
from ..utils.trackmate_spot import TrackMateSpot
from ..utils.tools import perform_cnn_inference
from ..utils.mitosis_track import MitosisTrack
from ..utils.cell_spot import CellSpot
from ..utils.trackmate_track import TrackMateTrack
from ..utils.trackmate_frame_spots import TrackMateFrameSpots
from ..utils.hidden_markov_models import HiddenMarkovModel
from ..utils.cell_division_detection.metaphase_cnn_model_params import (
    MetaphaseCnnModelParams,
)


def get_track_from_id(tracks: list[CellTrack], track_id: int) -> CellTrack:
    """
    Used to get track from its id.
    """
    for track in tracks:
        if track.track_id == track_id:
            return track
    raise ValueError(f"Track {track_id} not found")


class TracksMergingFactory:
    """
    Class to merge TrackMate tracks into mitosis tracks.

    Args:
        min_track_spots (int): Minimum spots in track to consider it.
        minimum_metaphase_interval (int): Minimum distance between two metaphases.
        max_spot_distance_for_split (int): Maximum distance between two spots to consider them.
    """

    def __init__(
        self,
        min_track_spots=10,
        minimum_metaphase_interval=10,
        max_spot_distance_for_split=20,
    ) -> None:
        self.min_track_spots = min_track_spots
        self.minimum_metaphase_interval = minimum_metaphase_interval
        self.max_spot_distance_for_split = max_spot_distance_for_split

    def read_trackmate_xml(
        self, xml_model_path: str, raw_video_shape: np.ndarray
    ) -> tuple[list[TrackMateTrack], list[TrackMateSpot]]:
        """
        Read useful information from xml file.
        """
        if not os.path.exists(xml_model_path):
            print("No xml file found for this video.")
            return None, None
        with open(xml_model_path) as fd:
            doc = xmltodict.parse(fd.read())

        # Define custom classes to read xml file
        trackmate_tracks = list(
            filter(
                lambda track: len(track.track_spots_ids) > 2
                and track.stop - track.start + 1 >= self.min_track_spots,
                [
                    TrackMateTrack(track)
                    for track in doc["TrackMate"]["Model"]["AllTracks"][
                        "Track"
                    ]
                ],
            )
        )
        raw_frames_spots = [
            TrackMateFrameSpots(spots, raw_video_shape)
            for spots in doc["TrackMate"]["Model"]["AllSpots"]["SpotsInFrame"]
        ]
        # Merge all frames - to get rid of TrackMateFrameSpots
        spots = [
            spot
            for raw_frame_spots in raw_frames_spots
            for spot in raw_frame_spots.spots
        ]

        return trackmate_tracks, spots

    def get_tracks_to_merge(
        self, raw_tracks: list[CellTrack]
    ) -> list[MitosisTrack]:
        """
        Plug tracks occurring at frame>0 to closest metaphase.
        """
        ordered_tracks = sorted(raw_tracks, key=lambda x: x.start)
        mitosis_tracks: list[MitosisTrack] = []

        # Loop through all tracks beginning at frame > 0 and try to plug them to the previous metaphase
        for track in reversed(ordered_tracks):
            # Break when reaching tracking starting at first frame, as they are ordered
            track_first_frame = min(track.spots.keys())
            if track_first_frame == 0:
                break

            # Get all spots at same frame
            contemporary_spots = [
                raw_track.spots[track_first_frame]
                for raw_track in raw_tracks
                if track_first_frame in raw_track.spots
                and raw_track.track_id != track.track_id
            ]

            # Keep only stuck spots
            first_spot = track.spots[track_first_frame]
            stuck_spots: list[CellSpot] = list(
                filter(
                    lambda x: x.is_stuck_to(
                        first_spot, self.max_spot_distance_for_split
                    ),
                    contemporary_spots,
                )
            )

            # Keep only spots with metaphase frame close to track first frame
            metaphase_spots = list(
                filter(
                    lambda x: get_track_from_id(
                        raw_tracks, x.track_id
                    ).has_close_metaphase(x, track_first_frame),
                    stuck_spots,
                )
            )

            # If no candidate has been found, ignore track
            if len(metaphase_spots) == 0:
                continue

            # Order remaining spots by overlap
            selected_spot = sorted(
                metaphase_spots,
                key=lambda x: get_track_from_id(
                    raw_tracks, x.track_id
                ).compute_metaphase_iou(track),
            )[-1]

            # Mother cell spot is spot in metaphase for corresponding track
            mother_cell_spot = selected_spot.corresponding_metaphase_spot

            # Check if it should be merged to existing split (division into 3 cells)
            triple_division = False
            for mitosis_track in mitosis_tracks:
                if mitosis_track.is_same_mitosis(
                    mother_cell_spot.track_id, mother_cell_spot.frame
                ):
                    mitosis_track.add_daughter_track(track.track_id)
                    triple_division = True
                    break

            # If not, create new split
            if not triple_division:
                mitosis_tracks.append(
                    MitosisTrack(
                        mother_cell_spot.track_id,
                        track.track_id,
                        mother_cell_spot.frame,
                    )
                )

        # Return dictionaries of tracks to merge
        return mitosis_tracks

    def correct_sequence(self, orig_sequence: list[int]) -> list[int]:
        """
        Correct sequence of states to fill the gap between two metaphase (1) subsequences
        separated by less than minimum_interval frames.

        Parameters
        ----------
        orig_seq: [class predicted]

        Returns
        -------
        seq: [class corrected]

        """
        corrected_sequence = np.copy(orig_sequence)
        # Get indexes of 1
        metaphase_index = [
            i for i, x in enumerate(corrected_sequence) if x == 1
        ]
        for idx in range(1, len(metaphase_index)):
            if (
                metaphase_index[idx] - metaphase_index[idx - 1] != 1
                and metaphase_index[idx] - metaphase_index[idx - 1]
                < self.minimum_metaphase_interval
            ):
                corrected_sequence[
                    list(range(metaphase_index[idx - 1], metaphase_index[idx]))
                ] = 1
        return corrected_sequence

    def pre_process_spots(
        self,
        trackmate_tracks: list[CellTrack],
        raw_spots: list[CellSpot],
        raw_video: np.array,
        metaphase_model_path: str,
        hmm_metaphase_parameters_file: str,
        predictions_file: str,
        video_name: str,
        only_predictions_update: bool,
    ) -> None:
        """
        Sort spots in track and predict metaphase.
        """

        nuclei_crops = []
        # Get list of possible metaphase spots
        for track in trackmate_tracks:
            # Get current track spots data & images
            current_nuclei_crops = track.get_spots_data(raw_spots, raw_video)
            # Merge current_nuclei_crops with nuclei_crops
            nuclei_crops = nuclei_crops + current_nuclei_crops  # CYX

        # Apply CNN model to get metaphase spots, once for all
        predictions = self._predict_metaphase_spots(
            metaphase_model_path, nuclei_crops
        )

        # Load HMM parameters and create model
        if not only_predictions_update:
            if not os.path.exists(hmm_metaphase_parameters_file):
                raise FileNotFoundError(
                    f"File {hmm_metaphase_parameters_file} not found"
                )
            hmm_parameters = np.load(hmm_metaphase_parameters_file)
            hmm_model = HiddenMarkovModel(
                hmm_parameters["A"], hmm_parameters["B"], hmm_parameters["pi"]
            )

        # Get list of possible metaphase spots
        for track in trackmate_tracks:
            track_predictions = predictions[: track.number_spots]
            predictions = predictions[track.number_spots :]

            # Hidden Markov Model to smooth predictions
            # If we just want to get raw CNN predictions, we don't want to correct the predictions
            if not only_predictions_update:
                track_predictions, _ = hmm_model.viterbi_inference(
                    track_predictions
                )
                track_predictions = self.correct_sequence(track_predictions)

            # Save prediction for each spot
            track.update_metaphase_spots(track_predictions)

        self.update_predictions_file(
            trackmate_tracks, predictions_file, video_name
        )

    @staticmethod
    def _predict_metaphase_spots(
        metaphase_model_path: str, nuclei_crops: list[np.array]
    ) -> list[int]:
        """
        Run CNN model to predict metaphase spots.

        Parameters
        ----------
        metaphase_model : str
            CNN model path
        nuclei_crops :  list[np.array]
            list[CYX]

        Returns
        -------
        predictions : list[int]
            predicted classes
        """

        predictions = perform_cnn_inference(
            model_path=metaphase_model_path,
            images=nuclei_crops,
            cnn_model_params=MetaphaseCnnModelParams,
        )
        return predictions

    @staticmethod
    def update_predictions_file(
        tracks: list[CellTrack], predictions_file: str, video_name: str
    ) -> None:
        """
        Parameters
        ----------
        tracks: [CellTrack]
        predictions_file: str
        video_name: str
        """
        if predictions_file is None:
            return

        # Read data from json prediction if exists
        if os.path.exists(predictions_file):
            with open(predictions_file) as json_file:
                predictions_data = json.load(json_file)
        else:
            predictions_data = {}

        # Retrieve predictions
        predictions = {
            int(track.track_id): [
                int(spot.predicted_phase) for spot in track.spots.values()
            ]
            for track in tracks
        }
        predictions_data[video_name] = predictions
        # Save predictions data
        with open(predictions_file, "w") as json_file:
            json.dump(predictions_data, json_file)



import os
from typing import Optional
import torch
import matplotlib.pyplot as plt
from cellpose import models

from cnn_framework.utils.readers.tiff_reader import TiffReader

from cut_detector.data.tools import get_data_path
from cut_detector.models.tools import get_model_path
from cut_detector.factories.segmentation_tracking_factory import (
    SegmentationTrackingFactory,
)


def main(
    image_path: Optional[str] = get_data_path("videos"),
    model_path: Optional[str] = os.path.join(
        get_model_path("segmentation"), "segmentation_model"
    ),
    diameter=0,  # 0 means using segmentation model saved value
    channel_to_segment=3,  # index starts at 1
    nucleus_channel=0,  # 0 means no nucleus channel
):
    """
    Script to run simple cellpose segmentation.
    """
    # If image_path or model_path are directories, take their first file
    if os.path.isdir(image_path):
        image_path = os.path.join(image_path, os.listdir(image_path)[0])
    if os.path.isdir(model_path):
        model_path = os.path.join(model_path, os.listdir(model_path)[0])

    # Read image and preprocess if needed
    image = TiffReader(image_path).image  # TCZYX

    # Initialize factory to get constants
    factory = SegmentationTrackingFactory(model_path)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model = models.CellposeModel(pretrained_model=[model_path], device=device)

    for frame in range(image.shape[0]):
        frame_image = image[frame, ...].squeeze()
        results, flows, _ = model.eval(
            [frame_image],
            channels=[channel_to_segment, nucleus_channel],
            diameter=diameter,
            flow_threshold=factory.flow_threshold,
            cellprob_threshold=factory.cellprob_threshold,
            augment=factory.augment,
        )

        # Plot image_to_segment and segmented_image
        _, ax = plt.subplots(2, 2)
        ax[0, 0].set_title("Raw image")
        ax[0, 0].imshow(frame_image[channel_to_segment - 1], cmap="gray")
        ax[0, 1].set_title("Cellpose")
        ax[0, 1].imshow(results[0], cmap="viridis")
        ax[1, 0].set_title("Flow")
        ax[1, 0].imshow(flows[0][0])
        ax[1, 1].set_title("Cell probability")
        ax[1, 1].imshow(flows[0][2], cmap="gray")
        plt.show()


if __name__ == "__main__":
    main()



import numpy as np


def zero_to_bottom_right(image):
    """
    Set to zero the bottom right part of the image.
    """
    image = np.copy(image)
    square_shape = image.shape[0]
    for i in range(square_shape):
        for j in range(square_shape):
            if i + j + 1 > square_shape:
                image[i, j] = 0
            if i + j + 1 == square_shape:  # diagonal
                image[i, j] = image[i, j] / 2
    return image


def zero_to_top_left(image):
    """
    Set to zero the top left part of the image.
    """
    image = np.copy(image)
    square_shape = image.shape[0]
    for i in range(square_shape):
        for j in range(square_shape):
            if i + j + 1 < square_shape:
                image[i, j] = 0
            if i + j + 1 == square_shape:  # diagonal
                image[i, j] = image[i, j] / 2
    return image


def zero_to_bottom_left(image):
    """
    Set to zero the bottom left part of the image.
    """
    image = np.copy(image)
    square_shape = image.shape[0]
    for i in range(square_shape):
        for j in range(square_shape):
            if i - j > 0:
                image[i, j] = 0
            if i - j == 0:  # diagonal
                image[i, j] = image[i, j] / 2
    return image


def zero_to_top_right(image):
    """
    Set to zero the top right part of the image.
    """
    image = np.copy(image)
    square_shape = image.shape[0]
    for i in range(square_shape):
        for j in range(square_shape):
            if i - j < 0:
                image[i, j] = 0
            if i - j == 0:  # diagonal
                image[i, j] = image[i, j] / 2
    return image


class MicroTubulesAugmentation:
    """
    Manage the augmentation of images for microtubules detection.
    """

    def __init__(self):
        self.augmentations = self.merge_augmentations()

    @classmethod
    def merge_augmentations(cls) -> dict[str, int]:
        """
        Merge values given to augmentation categories by different peaks.
        """
        augmentations = {}
        for category in [
            "top",
            "bottom",
            "left",
            "right",
            "top_left",
            "bottom_right",
            "top_right",
            "bottom_left",
        ]:
            value = 0  # by default, no MT is seen
            if value is not None:
                augmentations[category] = value
        return augmentations

    def generate_augmentations(
        self, original_image
    ) -> dict[str, dict[int, np.ndarray]]:
        """
        Generate image augmentations from the given image.

        Parameters
        ----------
        image: np.ndarray
            The image to augment. CYX

        """
        image = np.moveaxis(original_image, 0, -1)  # YXC
        assert image.shape[0] == image.shape[1]

        augmentations = {}

        if "top" in self.augmentations:
            transformed_image = np.copy(image)
            top_image = image[: image.shape[0] - image.shape[0] // 2, :]
            reversed_top_image = top_image[::-1, :]
            transformed_image[image.shape[0] // 2 :, :] = reversed_top_image
            augmentations["top"] = {
                "image": transformed_image,
                "category": self.augmentations["top"],
            }

        if "bottom" in self.augmentations:
            transformed_image = np.copy(image)
            bottom_image = image[image.shape[0] // 2 :, :]
            reversed_bottom_image = bottom_image[::-1, :]
            transformed_image[: image.shape[0] - image.shape[0] // 2, :] = (
                reversed_bottom_image
            )
            augmentations["bottom"] = {
                "image": transformed_image,
                "category": self.augmentations["bottom"],
            }

        if "left" in self.augmentations:
            transformed_image = np.copy(image)
            left_image = image[:, : image.shape[1] - image.shape[1] // 2]
            reversed_left_image = left_image[:, ::-1]
            transformed_image[:, image.shape[1] // 2 :] = reversed_left_image
            augmentations["left"] = {
                "image": transformed_image,
                "category": self.augmentations["left"],
            }

        if "right" in self.augmentations:
            transformed_image = np.copy(image)
            right_image = image[:, image.shape[1] // 2 :]
            reversed_right_image = right_image[:, ::-1]
            transformed_image[:, : image.shape[1] - image.shape[1] // 2] = (
                reversed_right_image
            )
            augmentations["right"] = {
                "image": transformed_image,
                "category": self.augmentations["right"],
            }

        if "top_left" in self.augmentations:
            transformed_image = zero_to_bottom_right(image) + np.moveaxis(
                zero_to_bottom_right(image)[::-1, ::-1].transpose(), 0, -1
            )
            augmentations["top_left"] = {
                "image": transformed_image,
                "category": self.augmentations["top_left"],
            }

        if "bottom_right" in self.augmentations:
            transformed_image = zero_to_top_left(image) + np.moveaxis(
                zero_to_top_left(image)[::-1, ::-1].transpose(), 0, -1
            )
            augmentations["bottom_right"] = {
                "image": transformed_image,
                "category": self.augmentations["bottom_right"],
            }

        if "top_right" in self.augmentations:
            transformed_image = zero_to_bottom_left(image) + np.moveaxis(
                zero_to_bottom_left(image).transpose(), 0, -1
            )
            augmentations["top_right"] = {
                "image": transformed_image,
                "category": self.augmentations["top_right"],
            }

        if "bottom_left" in self.augmentations:
            transformed_image = zero_to_top_right(image) + np.moveaxis(
                zero_to_top_right(image).transpose(), 0, -1
            )
            augmentations["bottom_left"] = {
                "image": transformed_image,
                "category": self.augmentations["bottom_left"],
            }

        return augmentations



import os
from typing import Optional

from ..factories.segmentation_tracking_factory import (
    SegmentationTrackingFactory,
)
from ..models.tools import get_model_path


def perform_tracking(
    video_path: str,
    fiji_path: str,
    save_folder: str,
    model_path: Optional[str],
    fast_mode: Optional[bool] = False,
) -> None:
    # Create save directory if it doesn't exist
    if not os.path.exists(save_folder):
        os.makedirs(save_folder)

    if model_path is None:
        model_path = os.path.join(
            get_model_path("segmentation"), "segmentation_model"
        )

    segmentation_tracking_factory = SegmentationTrackingFactory(model_path)
    segmentation_tracking_factory.perform_trackmate_tracking(
        video_path,
        fiji_path,
        save_folder,
        fast_mode,
    )



import os
from typing import Optional
from bigfish import stack

from cut_detector.widget_functions.mid_body_detection import (
    perform_mid_body_detection,
)
from cut_detector.data.tools import get_data_path


def main(
    data_set_path: Optional[str] = None,
):
    """
    If data_set_path is not None, it has to be a path containing
    folders: tracks, mitoses and videos.
    """
    if data_set_path is None:
        tracks_folder = get_data_path("tracks")
        mitoses_folder = get_data_path("mitoses")
        video_folder = get_data_path("videos")
    else:
        tracks_folder = os.path.join(data_set_path, "tracks")
        mitoses_folder = os.path.join(data_set_path, "mitoses")
        video_folder = os.path.join(data_set_path, "videos")

    video_files = os.listdir(video_folder)
    for video_index, video_file in enumerate(video_files):
        video_progress = f"{video_index + 1}/{len(video_files)}"
        print(f"\nVideo {video_progress}...")

        # Read video
        raw_video = stack.read_image(
            os.path.join(video_folder, video_file), sanity_check=False
        )  # TYXC
        video_name = os.path.basename(video_file).split(".")[0]

        perform_mid_body_detection(
            raw_video,
            video_name,
            mitoses_folder,
            tracks_folder,
        )


if __name__ == "__main__":
    main()
